"""Class definition for managing daemon config.
"""
import shutil
import configparser
from pathlib import Path


class DaemonConfig(object):
    """Config abstraction.

    Allows the creation, interpretation, and updating of a config file.

    Attributes
    ----------
    validator_ad_id_list : list
        Validator ad smart contract ID.
    validator_manager_mnemonic : str
        Validator mnemonic.
    algod_config_server : _type_, optional
        Algod URL, by default 'http://localhost:4001'.
    algod_config_token : str, optional
        Algod token, by default 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'.
    # indexer_config_server : _type_, optional
    #     Indexer URL, by default 'http://localhost:8980'.
    # indexer_config_token : str, optional
    #     Indexer token, by default 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'.
    # kmd_config_server : _type_, optional
    #     KMD URL, by default 'http://localhost:4002'.
    # kmd_config_token : str, optional
    #     KMD token, by default 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'.
    loop_period_s : int, optional
        Execution loop period in seconds, by default 3.
    # logging_level : str, optional
    #     Logging amount / level, by default 'DEBUG'.
    max_log_file_size_B : str, optional
        Maximal size of individual log files in bytes, by default 40*1024 (40 kB).
    num_of_log_files_per_level : str, optional
        Number of log files for each log level, by default 5.
    config_path : Path
        The config file location.
    config_filename : str
        Name of the config file.
    config_full_path : Path
        Full path to the config, including the filename.
    swap_full_path : Path
        Full path to the swap, including the filename.

    Methods
    -------
    get_swap_filemane
        Get the swap (temporary) filename.
    create_swap
        Copy the config to the same directory as a `.<...>.swp` file.
    read_config
        Read the configuration file, updating the class' parameters.
    read_swap
        Read swap file, updating the class' parameters.
    read
        Read daemon config file, updating the class' parameters.
    update_config
        Update config parameters.
    write_config
        Make config string and write it to a file.
    """


    def __init__(
        self,
        config_path: Path,
        config_filename: str,
    ):
        """Make config object.

        Parameters
        ----------
        config_path : Path
            The config file location.
        config_filename : str
            Name of the config file.
        """
        # Configuration parameters
        self.validator_ad_id_list = None
        self.validator_manager_mnemonic = None
        self.algod_config_server = None
        self.algod_config_token = None
        # self.indexer_config_server = None
        # self.indexer_config_token = None
        # self.kmd_config_server = None
        # self.kmd_config_token = None
        self.loop_period_s = None
        # self.logging_level = None
        # File paths and names
        self.config_path = config_path
        self.config_filename = config_filename
        self.config_full_path = Path(config_path, config_filename)
        self.swap_full_path = Path(config_path, self.get_swap_filename())

    
    def get_swap_filename(
        self
    ) -> str:
        """Get the swap (temporary) filename.

        Returns
        -------
        str
            Swap filename.
        """
        return f'.{self.config_filename}.swp'


    def create_swap(
        self
    ) -> None:
        """Copy the config to the same directory as a `.<...>.swp` file.
        """
        shutil.copy(
            self.config_full_path, 
            self.swap_full_path
        )
    

    def read_config(
        self
    ) -> None:
        """Read the configuration file, updating the class' parameters.
        """
        self.read(self.config_full_path)


    def read_swap(
        self
    ):
        """Read swap file, updating the class' parameters.
        """
        self.read(self.swap_full_path)


    def read(
        self,
        full_path: Path,
    ) -> None:
        """Read daemon config file, updating the class' parameters.

        Parameters
        ----------
        full_path : Path
            Path, including filename to the config file.

        Raises
        ------
        ValueError
            non-existent config.
        """
        # Catch non-existent file
        if not self.config_full_path.is_file():
            raise ValueError(f'Can\'t find the provided config file at {str(full_path)}')
        
        config = configparser.RawConfigParser(defaults=None, strict=False, allow_no_value=True)
        config.read(full_path)

        self.validator_manager_mnemonic = str(config.get('validator_config', 'validator_manager_mnemonic'))
        self.validator_ad_id_list = eval(config.get('validator_config', 'validator_ad_id_list'))

        self.algod_config_server =   str(config.get('algo_client_config', 'algod_config_server'))
        self.algod_config_token =    str(config.get('algo_client_config', 'algod_config_token'))
        # self.indexer_config_server = str(config.get('algo_client_config', 'indexer_config_server'))
        # self.indexer_config_token =  str(config.get('algo_client_config', 'indexer_config_token'))
        # self.kmd_config_server =     str(config.get('algo_client_config', 'kmd_config_server'))
        # self.kmd_config_token =      str(config.get('algo_client_config', 'kmd_config_token'))

        # self.loop_period_s = float(config.get('daemon_config', 'loop_period_s'))
        # self.logging_level = str(config.get('daemon_config', 'logging_level')).upper()

        self.max_log_file_size_B = int(eval(config.get('logging_config', 'max_log_file_size_B')))
        self.num_of_log_files_per_level = int(eval(config.get('logging_config', 'num_of_log_files_per_level')))

        self.loop_period_s = float(config.get('runtime_config', 'loop_period_s'))


    def update_config(
        self,
        validator_ad_id_list: list,
        validator_manager_mnemonic: str,
        algod_config_server: str='http://localhost:4001',
        algod_config_token: str='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
        # indexer_config_server: str='http://localhost:8980',
        # indexer_config_token: str='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
        # kmd_config_server: str='http://localhost:4002',
        # kmd_config_token: str='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
        loop_period_s: int=15,
        # logging_level: str='DEBUG'
        max_log_file_size_B: int=40*1024,
        num_of_log_files_per_level: int=3
    ) -> None:
        """Update config parameters.

        Parameters
        ----------
        validator_ad_id_list : list
            Validator ad smart contract ID.
        validator_manager_mnemonic : str
            Validator mnemonic.
        algod_config_server : _type_, optional
            Algod URL, by default 'http://localhost:4001'.
        algod_config_token : str, optional
            Algod token, by default 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'.
        # indexer_config_server : _type_, optional
        #     Indexer URL, by default 'http://localhost:8980'.
        # indexer_config_token : str, optional
        #     Indexer token, by default 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'.
        # kmd_config_server : _type_, optional
        #     KMD URL, by default 'http://localhost:4002'.
        # kmd_config_token : str, optional
        #     KMD token, by default 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'.
        loop_period_s : int, optional
            Execution loop period in seconds, by default 15.
        # logging_level : str, optional
        #     Logging amount / level, by default 'DEBUG'.
        max_log_file_size_B : str, optional
            Maximal size of individual log files in bytes, by default 40*1024 (40 kB).
        num_of_log_files_per_level : str, optional
            Number of log files for each log level, by default 3.
        """
        self.validator_ad_id_list = validator_ad_id_list
        self.validator_manager_mnemonic = validator_manager_mnemonic
        self.algod_config_server = algod_config_server
        self.algod_config_token = algod_config_token
        # self.indexer_config_server = indexer_config_server
        # self.indexer_config_token = indexer_config_token
        # self.kmd_config_server = kmd_config_server
        # self.kmd_config_token = kmd_config_token
        self.loop_period_s = loop_period_s
        # self.logging_level = logging_level
        self.max_log_file_size_B = max_log_file_size_B
        self.num_of_log_files_per_level = num_of_log_files_per_level


    def write_config(
        self,
    ) -> None:
        """Make config string and write it to a file.
        """
        config_content_string = '\n' + \
        '[validator_config] #####################################################################################################' + '\n' + \
        '\n' + \
        f'validator_ad_id_list = {self.validator_ad_id_list}' + '\n' + \
        f'validator_manager_mnemonic = {self.validator_manager_mnemonic}' + '\n' + \
        '\n\n' + \
        '[algo_client_config] ###################################################################################################' + '\n' + \
        '\n' + \
        f'algod_config_server = {self.algod_config_server}' + '\n' + \
        f'algod_config_token = {self.algod_config_token}' + '\n' + \
        '\n\n' + \
        '[logging_config] #######################################################################################################' + '\n' + \
        '\n' + \
        f'max_log_file_size_B = {self.max_log_file_size_B}' + '\n' \
        f'num_of_log_files_per_level = {self.num_of_log_files_per_level}' + '\n' \
        '\n\n' + \
        '[runtime_config] #######################################################################################################' + '\n' + \
        '\n' + \
        f'loop_period_s = {self.loop_period_s}' + '\n'

        with open(self.config_full_path, 'w') as f:
            f.write(config_content_string)



"""App wrapper definitions for caching the state of contracts and keeping reference to the corresponding clients.
"""
import logging
from typing import List, Tuple
from urllib.error import URLError

from algosdk.encoding import encode_address
from algokit_utils.beta.algorand_client import AlgorandClient
from algosdk.error import AlgodHTTPError

from valar_daemon.Logger import Logger
from valar_daemon.ValidatorAdClient import (
    ValidatorAdClient,
    GlobalState as ValidatorAdGlobalState
)
from valar_daemon.DelegatorContractClient import (
    DelegatorContractClient,
    GlobalState as DelegatorContractGlobalState
)
from valar_daemon.NoticeboardClient import NoticeboardClient
from valar_daemon.utils import (
    get_delco_fee_and_gating_asa_id,
    decode_delegation_terms_general,
    decode_uint64_list
)
from valar_daemon.constants import (
    VALAD_STATE_DELETED_MASK,
    DELCO_STATE_DELETED_MASK
)


class AppWrapper(object):
    """Smart contract (app) wrapper."""

    def __init__(self):
        pass


class ValadAppWrapper(AppWrapper):
    """Validator Ad wrapper.

    Attributes
    ----------
    algorand_client: AlgorandClient
        Algorand client.
    app_id: int
        Validator Ad application ID.
    valad_client: ValidatorAdClient
        Validator ad client.
    notbd_client: NoticeboardClient
        Noticeboard client.
    valown_address: str
        Address of the validator owner.
    delco_id_list: list
        [dynamic] List of the IDs of connected delegator contracts.
    state: bytes
        [dynamic] State of the validator ad.
    """
    def __init__(
        self,
        algorand_client: AlgorandClient,
        app_id: int
    ):
        """Validator Ad wrapper, used for storing the client and relevant parameters for the daemon's operation.

        Parameters
        ----------
        algorand_client : AlgorandClient
            Algorand client.
        app_id : int
            Validator Ad ID.
        """
        self.app_id = app_id
        self.valad_client = ValidatorAdClient(
            algorand_client.client.algod,
            app_id=app_id
        )
        # Fetch global state and populate corresponding atributes
        valad_global_state = self.update_dynamic(propagate_deleted_error=True)
        self.notbd_client = NoticeboardClient(
            algorand_client.client.algod,
            app_id=valad_global_state.noticeboard_app_id
        )
        self.notbd_client.get_global_state() # Try to see if it can be connected to (error caught in upper layers)
        self.valown_address = encode_address(valad_global_state.val_owner.as_bytes)

    def update_dynamic(
            self,
            propagate_deleted_error: bool=True
        ) -> object:
        """Update the dynamic parts of the valad app.

        Notes
        -----
        Can handle non-existent error and set the state accordingly.

        Parameters
        ----------
        propagate_deleted_error : bool, optional
            If true, does not handle any errors that are raised, by default True

        Returns
        -------
        object
            The app's global state.

        Raises
        ------
        e
            Given error type.
        """
        try:
            valad_global_state = self.valad_client.get_global_state()
        except AlgodHTTPError as e:
            # Update to show the thing was deleted
            if not propagate_deleted_error and e.code == 404 and e.args[0] == 'application does not exist':
                self.state = VALAD_STATE_DELETED_MASK
                return None
            else:
                raise e
        except Exception as e: # Must be last to capture AlgodHTTPError first and execute corresponding logic
            raise e
        self.delco_id_list = ValadAppWrapper.get_delco_id_list(valad_global_state)
        self.state = valad_global_state.state.as_bytes
        return valad_global_state

    @staticmethod
    def get_delco_id_list(
        valad_global_state: ValidatorAdGlobalState
    ):
        """Get list of IDs of delcos that are associated with the valad (between, including, ready and live state).

        Parameters
        ----------
        valad_global_state : ValidatorAdGlobalState
            Get the global state of the valad.

        Returns
        -------
        List
            IDs of the associated delcos.
        """
        delco_id_list = decode_uint64_list(
            # valad_app.valad_client.get_global_state().del_app_list.as_bytes
            valad_global_state.del_app_list.as_bytes
        )
        return [i for i in delco_id_list if i != 0]


class DelcoAppWrapper(AppWrapper):
    """Delegator Contract wrapper.

    Attributes
    ----------
    app_id: int
        Delegator Contract application ID.
    delco_client: DelegatorContractClient
        Delegator contract client.
    notbd_client: NoticeboardClient
        Noticeboard client.
    valad_id: int
        Validator Ad app ID.
    delman_address: str
        Delegator manager address.
    delben_address: str
        Delegator beneficiary address.
    fee_asa_id: int
        Asset ID for issuing payment.
    gating_asa_id_list: List[Tupe(int, int)]
        Garing asset IDs and their minimal values.
    round_start: int
        Staking start (partkey validity).
    round_end: int
        Staking end (partkey validity).
    valown_address: int
        Validator owner address.
    partner_address: int
        Address of the partner that forwarded the delegator.
    round_ended: int
        [dynamic] Round at which the contract ended.
    state: bytes
        [dynamic] State of the validator ad.
    """
    def __init__(
        self,
        algorand_client,
        app_id: int
    ):
        """Delegator Contract wrapper, used for storing the client and relevant parameters for the daemon's operation.

        Parameters
        ----------
        algorand_client : AlgorandClient
            Algorand client.
        app_id : int
            Delegator Contract ID.
        """
        self.app_id = app_id
        self.delco_client = DelegatorContractClient(
            algorand_client.client.algod,
            app_id=app_id
        )
        delco_global_state = self.update_dynamic(propagate_deleted_error=True)
        self.notbd_client = NoticeboardClient(
            algorand_client.client.algod,
            app_id=delco_global_state.noticeboard_app_id
        )
        self.notbd_client.get_global_state() # Try to see if it can be connected to (error caught in upper layers)
        self.valad_id = delco_global_state.validator_ad_app_id
        self.delman_address = encode_address(delco_global_state.del_manager.as_bytes)
        self.delben_address = encode_address(delco_global_state.del_beneficiary.as_bytes)
        self.fee_asa_id, self.gating_asa_id_list = get_delco_fee_and_gating_asa_id(delco_global_state)
        self.round_start = delco_global_state.round_start
        self.round_end = delco_global_state.round_end
        # self.round_end = delco_global_state.round_end + 320 # Account for initial 320 round delay
        # self.round_key_delete = self.round_end # Default value - contract end and key deletion can be earlier
        valad_client = ValidatorAdClient(
                algorand_client.client.algod,
                app_id=delco_global_state.validator_ad_app_id
            )
        valad_global_state = valad_client.get_global_state()
        self.valown_address = encode_address(valad_global_state.val_owner.as_bytes)
        self.partner_address = decode_delegation_terms_general(
            delco_global_state.delegation_terms_general.as_bytes
        ).partner_address

    def update_dynamic(
            self,
            propagate_deleted_error: bool=True
        ) -> DelegatorContractGlobalState:
        """Update the dynamic parts of the delco app.

        Notes
        -----
        Can handle non-existent error and set the state accordingly.

        Parameters
        ----------
        propagate_deleted_error : bool, optional
            If true, does not handle any errors that are raised, by default True

        Returns
        -------
        DelegatorContractGlobalState
            The app's global state.

        Raises
        ------
        e
            Given error type.
        """
        try:
            delco_global_state = self.delco_client.get_global_state()
        except AlgodHTTPError as e:
            # Update to show the thing was deleted
            if not propagate_deleted_error and e.code == 404 and e.args[0] == 'application does not exist':
                self.state = DELCO_STATE_DELETED_MASK
                return None
            else:
                raise e
        except Exception as e:
            raise e
        self.state = delco_global_state.state.as_bytes
        self.round_ended = delco_global_state.round_ended
        return delco_global_state


class AppWrapperList(object):
    """List of app wrappers.

    Attributes
    ----------
    algorand_client : AlgorandClient
        Algorand client.
    logger : logging.Logger
        Logger.
    AppWrapperClass: AppWrapper
        Either `ValadAppWrapper` or `DelcoAppWrapper`.
    """
    def __init__(
        self,
        algorand_client: AlgorandClient,
        logger: Logger,
        AppWrapperClass: AppWrapper
    ):
        """List of app wrappers.

        Parameters
        ----------
        algorand_client : AlgorandClient
            Algorand client.
        AppWrapperClass : AppWrapper
            Either `ValadAppWrapper` or `DelcoAppWrapper`.
        """
        self.algorand_client = algorand_client
        self.logger = logger
        self.AppWrapperClass = AppWrapperClass
        self.app_list = []

    def get_app_list(
        self
    ) -> List[AppWrapper]:
        """Return the app list.

        Returns
        -------
        list
        """
        return self.app_list

    def get_id_list(
        self
    ) -> List[int]:
        """Get the IDs of the apps in the list.

        Returns
        -------
        List[int]
        """
        return [app.app_id for app in self.app_list]

    def add_single_app(
        self,
        app_id: int
    ):
        """Add an app to the list.

        Parameters
        ----------
        app_id : int
            ID of the app to add.

        Returns
        -------
        int
            Number of added apps (0 or 1).
        """
        app_wrapper = None
        # Try to connect the client and get state; log error if not successful
        try:
            app_wrapper = self.AppWrapperClass(self.algorand_client, app_id)
        except URLError as e:
            # self.logger.critical(f'For app ID {app_id}, URLError {e.args[0].errno}: {e.args[0].strerror}.')
            self.logger.log_app_create_urlerror(
                app_id=app_id, errno=e.args[0].errno, strerror=e.args[0].strerror
            )
            return 0
        except AlgodHTTPError as e:
            # self.logger.critical(f'AlgodHTTPError {e.code}: {e.args[0]}.')
            self.logger.log_app_create_algohttperror(
                app_id=app_id, errno=e.code, strerror=e.args[0]
            )
            return 0
        except Exception as e:
            # self.logger.error(e)
            self.logger.log_app_create_genericerror(app_id=app_id, e=e)
            return 0
        # Add app wrapper if successfully initialized
        self.app_list.append(app_wrapper)
        return 1

    def remove_single_app(
        self,
        app_id: int
    ) -> bool:
        """Remove an app from the list.

        Parameters
        ----------
        app_id : int

        Returns
        -------
        bool
            Flag that is `True` if the app was found and deleted.
        """
        for i, app in enumerate(self.app_list):
            if app.app_id == app_id:
                self.app_list.pop(i)
                return True
        return False

    def add_multiple_apps(
        self,
        app_id_list: List[int]
    ) -> int:
        """Add multiple apps to the list.

        Parameters
        ----------
        app_id_list : List[int]
            IDs of apps to add.

        Returns
        -------
        int
            Number of added apps.
        """
        existing_app_ids = self.get_id_list()
        num_of_added_apps = 0
        for app_id in app_id_list:
            if app_id not in existing_app_ids: # Only add new apps
                num_of_added_apps += self.add_single_app(app_id)
        return num_of_added_apps

    def remove_multiple_apps(
        self,
        app_id_list: List[int]
    ):
        """Remove multiple apps from the list.

        Parameters
        ----------
        app_id_list : List[int]
        """
        for app_id in app_id_list:
            self.remove_single_app(app_id)

    def update_apps(
        self
    ) -> Tuple[int, int]:
        """Update the dynamic parts of the connected apps.
        """
        num_of_apps = len(self.app_list)
        num_of_updated = 0
        for app in self.app_list:
            try:
                app.update_dynamic(propagate_deleted_error=False) # Assign 'deleted' state if exists
                num_of_updated += 1
            except Exception as e:
                # self.logger.error(e)
                self.logger.log_app_dynamic_update_genericerror(app_id=app.app_id, e=e)
        return (num_of_apps, num_of_updated)


class ValadAppWrapperList(AppWrapperList):
    """List of Validator Ad app wrappers.

    Attributes
    ----------
    algorand_client: AlgorandClient
        Algorand client.
    """
    def __init__(
        self,
        algorand_client: AlgorandClient,
        logger: logging.Logger
    ):
        """List of Validator Ad app wrappers.

        Parameters
        ----------
        algorand_client : AlgorandClient
            Algorand client.
        logger : logging.Logger
            Logger.
        """
        super().__init__(
            algorand_client=algorand_client,
            logger=logger,
            AppWrapperClass=ValadAppWrapper
        )


class DelcoAppWrapperList(AppWrapperList):
    """List of Delegator Contract app wrappers.

    Attributes
    ----------
    algorand_client : AlgorandClient
        Algorand client.
    logger : logging.Logger
        Logger.
    """
    def __init__(
        self,
        algorand_client: AlgorandClient,
        logger: logging.Logger
    ):
        """List of Delegator Contract app wrappers.

        Parameters
        ----------
        algorand_client : AlgorandClient
            Algorand client.
        """
        super().__init__(
            algorand_client=algorand_client,
            logger=logger,
            AppWrapperClass=DelcoAppWrapper
        )



"""Various constants.
"""

### Daemon-specific

"""
Delegator contract ready sub-state constants, indicating:

NOT_SUBMITTED   - Time for submitting the partkeys according to the validator ad's terms is up.
REQUEST_DENIED  - Requesting partkey generation was not succesfull.
BREACH_LIMITS   - Limits breached.
BREACH_PAY      - Payment breached.
UNKNOWN_ERROR   - An unknown error occurred while submitting partkeys.
URL_ERROR       - A known URL error (no. 111 or -2) occurred while submitting partkeys.
REQUESTED       - Partkey generation was requested.
PENDING         - Partkey generation is pending.
SUBMITTED       - Partkeys were submitted and may update the contract state.
"""
DELCO_READY_STATUS_NOT_SUBMITTED    = -6
DELCO_READY_STATUS_REQUEST_DENIED   = -5
DELCO_READY_STATUS_BREACH_LIMITS    = -4
DELCO_READY_STATUS_BREACH_PAY       = -3
DELCO_READY_STATUS_UNKNOWN_ERROR    = -2
DELCO_READY_STATUS_URL_ERROR        = -1
DELCO_READY_STATUS_REQUESTED        =  0
DELCO_READY_STATUS_PENDING          =  1
DELCO_READY_STATUS_SUBMITTED        =  2

"""
Validator ad not ready sub-state constants, indicating:

VALAD_NOT_READY_STATUS_NO_CHANGE        - No change needed, no error recorded.
VALAD_NOT_READY_STATUS_CHANGE_OK        - Switch from not ready to ready successful.
VALAD_NOT_READY_STATUS_ATTRIBUTE_ERROR  - AttributeError during switch, likely due to broken algod connection.
"""
VALAD_NOT_READY_STATUS_NO_CHANGE =  1
VALAD_NOT_READY_STATUS_CHANGE_OK =  0
VALAD_NOT_READY_STATUS_ATTRIBUTE_ERROR = -1


### From the smart contracts

ZERO_ADDRESS = "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAY5HFKQ"
"""str: algorand encoded address of 32 zero bytes"""

ROLE_VAL = b"val_"
ROLE_DEL = b"del_"
"""
Possible user roles (i.e. prefix for user's box at Noticeboard):
    ROLE_VAL - user is a validator.
    ROLE_DEL - user is a delegator.
"""

BOX_ASA_KEY_PREFIX = b"asa_"
"""
BOX_ASA_KEY_PREFIX : bytes
    Prefix for the boxes of ASA at ValidatorAd.
"""

ALGO_ASA_ID = 0
"""ID of Algo ASA."""

"""
Possible states of the contract:
    CREATED - validator ad has been created.
    TEMPLATE_LOAD - validator ad is getting loaded the delegator contract template.
    TEMPLATE_LOADED - validator ad ended loading of the delegator contract template.
    SET - initial terms of validator ad have been set.
    READY - validator ad manager is ready to accept new delegators.
    NOT_READY - validator ad manager is not ready to accept new delegators.
    NOT_LIVE - validator ad owner does not want to accept new delegators.
    DELETED - fetching the state of a connected client raises a 'non-existent' error.
"""
VALAD_STATE_NONE = b"\x00"
VALAD_STATE_CREATED = b"\x01"
VALAD_STATE_TEMPLATE_LOAD = b"\x02"
VALAD_STATE_TEMPLATE_LOADED = b"\x03"
VALAD_STATE_SET = b"\x04"
VALAD_STATE_READY = b"\x05"
VALAD_STATE_NOT_READY = b"\x06"
VALAD_STATE_NOT_LIVE = b"\x07"
VALAD_STATE_DELETED_MASK = b"\x20"

"""
Possible states of the contract:
    CREATED - contract has been created.
    LIVE - contract is live.
    READY - waiting for keys submission.
    SET - contract terms have been set.
    SUBMITTED - waiting for keys confirmation.
    ENDED_NOT_SUBMITTED - keys have not been submitted in time.
    ENDED_NOT_CONFIRMED - keys have not been confirmed in time.
    ENDED_LIMITS - maximum number of limit breach events has been reached.
    ENDED_WITHDREW - delegator withdrew from the contract prematurely.
    ENDED_EXPIRED - contract has expired.
    ENDED_UPTIME - validator has breach the agreed uptime guarantee.
    ENDED_CANNOT_PAY - delegator cannot pay the validator (as funds could have been frozen and/or clawed back).
    DELETED - fetching the state of a connected client raises a 'non-existent' error.
"""
DELCO_STATE_NONE = b"\x00"
DELCO_STATE_CREATED = b"\x01"
DELCO_STATE_SET = b"\x02"
DELCO_STATE_READY = b"\x03"
DELCO_STATE_SUBMITTED = b"\x04"
DELCO_STATE_LIVE = b"\x05"
DELCO_STATE_ENDED_NOT_SUBMITTED = b"\x10"
DELCO_STATE_ENDED_NOT_CONFIRMED = b"\x11"
DELCO_STATE_ENDED_LIMITS = b"\x12"
DELCO_STATE_ENDED_WITHDREW = b"\x13"
DELCO_STATE_ENDED_EXPIRED = b"\x14"
DELCO_STATE_ENDED_UPTIME = b"\x15"
DELCO_STATE_ENDED_CANNOT_PAY = b"\x16"
DELCO_STATE_ENDED_MASK = b"\x10"
DELCO_STATE_DELETED_MASK = b"\x20"
